"use strict";

var url = require('url');
var rq = require('request');

module.exports.createRestClient = function (urlBase) {
    return {
        urlBase: urlBase,

        /**
         * Retrieves the client message with the specified double-sha256 hash
         * @param messageHash A buffer of 32 bytes
         * @param callback parameters (error, serializedMessage) if first
         * parameter is not null, an error occurred.
         * If first parameter is null, then the second parameter is a buffer
         * with the serialized client message. If no such client message exists,
         * the callback will be called with (null, null).
         */
        getTransaction: function (messageHash, callback) {
            validateMessageHash(messageHash)
            get(urlBase, 'tx/' + messageHash.toString('hex'), (error, statusCode, responseObject) => {
                handleGetResponse(error, statusCode, Buffer.from(responseObject.tx, 'hex'), callback);
            });

        },

        postTransaction: function (serializedClientMessage, callback) {
            if (!Buffer.isBuffer(serializedClientMessage)) {
                throw new Error("messageHash is not a Buffer");
            }
            var jsonObject = {tx: serializedClientMessage.toString('hex')};
            doPost(urlBase, 'tx', jsonObject, callback);
        },

        getConfirmationProof: function (messageHash, callback) {
            validateMessageHash(messageHash);
            get(urlBase, 'tx/' + messageHash.toString('hex') + '/confirmationProof', (error, statusCode, responseObject) => {
                handleGetResponse(error, statusCode, responseObject, callback);
            });
        },

        status: function (messageHash, callback) {
            validateMessageHash(messageHash);
            get(urlBase, 'tx/' + messageHash.toString('hex') + '/status', (error, statusCode, responseObject) => {
                handleGetResponse(error, statusCode, responseObject, callback);
            });
        },

        query: function(queryObject, callback) {
            doPost(urlBase, 'query/', queryObject, (error, result) => {
                if (error) {
                    callback(error);
                    return;
                }
                callback(null, result);
            });
        }
    };
}

function doPost(urlBase, path, jsonObject, responseCallback) {
    post(urlBase, path, jsonObject, (error, statusCode, responseObject) => {
        if (error) {
            console.log("In restclient doPost().", error);
            responseCallback(error);
        } else if (statusCode != 200) {
            console.log(`Unexpected status code from server: ${statusCode}`);
            responseCallback(new Error("Unexpected status code from server: " + statusCode));
        } else {
            try {
                console.log(`Ok calling responseCallback with responseObject: ${JSON.stringify(responseObject)}`);
                responseCallback(null, responseObject);
            } catch (error) {
                console.log(`error calling responseCallback: ${error}`);
                responseCallback(error);
            }
        }
    });
}

function get(urlBase, path, callback) {
    console.log("GET URL " + url.resolve(urlBase, path));
  rq(url.resolve(urlBase, path), function (err, response, body) {
        if (err) callback(err);
        else {
            try {
                callback(null, response.statusCode, JSON.parse(body));
            } catch (e) {
                callback(e);
            }
        }
    });
}

function post(urlBase, path, jsonBody, callback) {
    console.log("POST URL " + url.resolve(urlBase, path));
  rq({method: 'POST', url: url.resolve(urlBase, path),
      json: true, body: jsonBody},
    function (err, response, body) {
        if (err) callback(err);
        else {
            try {
                callback(null, response.statusCode, body);
            } catch (e) {
                callback(e);
            }
        }
    })
     
}

function validateMessageHash(messageHash) {
    if (!Buffer.isBuffer(messageHash)) {
        throw new Error("messageHash is not a Buffer");
    }
    if (messageHash.length != 32) {
        throw new Error("expected length 32 of messageHash, but got " + messageHash.length);
    }
}

// TODO: change order of arguments
function handleGetResponse(error, statusCode, responseObject, callback) {
    if (error) {
        callback(error);
    } else if (statusCode === 404) {
        callback(null, null);
    } else if (statusCode != 200) {
        callback(new Error("Unexpected status code from server: " + statusCode));
    } else {
        try {
            callback(null, responseObject);
        } catch (error) {
            callback(error);
        }
    }
}

